/*////////////////////////////////////////////////////////////////////

 Author : Crusader Mike

 Purpose : this file contains template classes that can 
           be used for building thunks for 'thiscall',
           'stdcall' and 'cdecl' calling conventions.

           Thunks are very suitable and efficient way to avoid tiresome
           TLS/CritSecs/maps (when providing apartment threading and reentrancy).

 Notes : thanks to Andrew Nosenko (andien@geocities.com) for 
         idea and thunks for 'thiscall' and 'stdcall'
 Using :
         CAuxThunk<> is provided for __thiscall methods,
         CAuxStdThunk<> is for __stdcall,
         CAuxCDeclThunk<> is for __cdecl
         CAuxCDeclThunk0<> is a specialized case for __cdecl thunking
         where thunked function has no parameters (this is because MSVC
         does not support partial template specialization)

Here is an example of how to set up Win32 hook the reentrant way, without a
use of global variable:
  
class CHook : public CAuxThunk<CHook>
{
public:
    CHook(): m_hhook(NULL)
    {
        InitThunk((TMFP)CBTHook, this);
    }

    BOOL Hook()
    {
        m_hook = SetWindowsHookEx(WH_CBT, (HOOKPROC)GetThunk(), NULL, GetCurrentThreadId());
        return (BOOL)m_hook;
    }

    LRESULT CBTHook(int nCode, WPARAM wParam, LPARAM lParam);

    HHOOK m_hhook;
...
};

LRESULT CHook::CBTHook(int nCode, WPARAM wParam, LPARAM lParam)
{
    if ( nCode == HCBT_CREATEWND )
    {
        UnhookWindowsHookEx(m_hook);
        HWND hwnd = (HWND)wParam;
        // do whatever we want with HWND
        ...
    }
    return CallNextHookEx(m_hook, nCode, wParam, lParam);
}

////////////////////////////////////////////////////////////////////*/

#ifndef THUNKS_H_
#define THUNKS_H_

#ifndef _M_IX86
    #error Thunks are implemented for X86 only!
#endif

#pragma pack(push, 1)

////////////////////////////////////////////////////////////////////
// CAuxThunk<>

template <class T>
class CAuxThunk
{
    BYTE    m_mov;          // mov ecx, %pThis
    DWORD   m_this;         //
    BYTE    m_jmp;          // jmp func
    DWORD   m_relproc;      // relative jmp

public:
    typedef void (T::*TMFP)();
    void InitThunk(TMFP method, const T* pThis)
    {
        union { DWORD func; TMFP method; } addr;
        addr.method = (TMFP)method;

        m_mov       = 0xB9;
        m_this      = (DWORD)pThis;
        m_jmp       = 0xE9;
        m_relproc   = addr.func - (DWORD)(this + 1);

        FlushInstructionCache(GetCurrentProcess(), this, sizeof(*this));
    }
    FARPROC GetThunk() const
    {
        _ASSERTE(m_mov == 0xB9);
        return (FARPROC)this;
    }
};

////////////////////////////////////////////////////////////////////
// CAuxStdThunk<>

template <class T>
class CAuxStdThunk
{
    BYTE    m_mov;          // mov eax, %pThis
    DWORD   m_this;         //
    DWORD   m_xchg_push;    // xchg eax, [esp] : push eax
    BYTE    m_jmp;          // jmp func
    DWORD   m_relproc;      // relative jmp

public:
    typedef void (__stdcall T::*TMFP)();
    void InitThunk(TMFP method, const T* pThis)
    {
        union { DWORD func; TMFP method; } addr;
        addr.method = method;

        m_mov       = 0xB8;
        m_this      = (DWORD)pThis;
        m_xchg_push = 0x50240487;
        m_jmp       = 0xE9;
        m_relproc   = addr.func - (DWORD)(this + 1);

        FlushInstructionCache(GetCurrentProcess(), this, sizeof(*this));
    }
    FARPROC GetThunk() const
    {
        _ASSERTE(m_mov == 0xB8);
        return (FARPROC)this;
    }
};

////////////////////////////////////////////////////////////////////
// CAuxCDeclThunk<>
// due to the nature of __cdecl calls it is necessary to know number 
// of arguments (nArgs) at this point.

template <class T, UINT nArgs>
class CAuxCDeclThunk
{
    BYTE    m_push_esi;     //0x56      // push esi
    BYTE    m_push_edi;     //0x57      // push edi
    BYTE    m_mov_ecx;      //0xB9      // mov ecx, nArgs*4
    DWORD   m_nArgs_x_4;
    WORD    m_mov_esi_esp;  //0xF48B    // mov esi, esp
    WORD    m_add_esi_12;   //0xC683    // add esi, 12
    BYTE    m_12;
    WORD    m_sub_esp_ecx;  //0xE12B    // sub esp, ecx
    WORD    m_mov_edi_esp;  //0xFC8B    // mov edi, esp
    WORD    m_repne_movsb;  //0xA4F2    // repne movsb

    BYTE    m_push;         //0x68      // push pThis 
    DWORD   m_this;
    BYTE    m_call;         //0xE8      // call relproc
    DWORD   m_relproc;

    WORD    m_add_esp_data; //0xC481    // add esp, nArgs*4 + 4
    DWORD   m_data;
    BYTE    m_pop_edi;      //0x5F      // pop edi
    BYTE    m_pop_esi;      //0x5E      // pop esi
    BYTE    m_ret;          //0xC3      // ret

public:
    CAuxCDeclThunk()
    {
        m_push_esi      = 0x56;
        m_push_edi      = 0x57;
        m_mov_ecx       = 0xB9;
        m_nArgs_x_4     = nArgs*4;
        m_mov_esi_esp   = 0xF48B;
        m_add_esi_12    = 0xC683;
        m_12            = 0x0C;
        m_sub_esp_ecx   = 0xE12B;
        m_mov_edi_esp   = 0xFC8B;
        m_repne_movsb   = 0xA4F2;

        m_push          = 0x68;
        m_this          = 0xFFFFFFFF;
        m_call          = 0xE8;
//      m_relproc;

        m_add_esp_data  = 0xC481;
        m_data          = nArgs*4 + 4;
        m_pop_edi       = 0x5F;
        m_pop_esi       = 0x5E;
        m_ret           = 0xC3;
    }               

    typedef void (__cdecl T::*TMFP)();

    void InitThunk(TMFP method, const T* pThis)
    {
        m_this      = (DWORD)pThis;

        union { DWORD func; TMFP method; } addr;
        addr.method = method;
        m_relproc   = addr.func - ((DWORD)(this) + 28);

        FlushInstructionCache(GetCurrentProcess(), this, sizeof(*this));
    }

    FARPROC GetThunk() const
    {
        _ASSERTE(m_this != 0xFFFFFFFF);
        return (FARPROC)this;
    }
};

////////////////////////////////////////////////////////////////////
// CAuxCDeclThunk0<>
// specialized case of __cdecl call with zero arguments

template <class T>
class CAuxCDeclThunk0
{
    BYTE    m_push;         //0x68  // push pThis
    DWORD   m_this;
    BYTE    m_call;         //0xE8  // call relproc
    DWORD   m_relproc;
    BYTE    m_pop;          //0x59  // pop ecx
    BYTE    m_ret;          //0xC3  // ret

public:
    CAuxCDeclThunk0()
    {
        m_push      = 0x68;
        m_this      = 0xFFFFFFFF;
        m_call      = 0xE8;
        m_pop       = 0x59;
        m_ret       = 0xC3;
    }

    typedef void (__cdecl T::*TMFP)();

    void InitThunk(TMFP method, const T* pThis)
    {
        m_this      = (DWORD)pThis;

        union { DWORD func; TMFP method; } addr;
        addr.method = method;
        m_relproc   = addr.func - ((DWORD)(this) + 10);

        FlushInstructionCache(GetCurrentProcess(), this, sizeof(*this));
    }

    FARPROC GetThunk() const
    {
        _ASSERTE(m_this != 0xFFFFFFFF);
        return (FARPROC)this;
    }
};

#pragma pack(pop) // CAuxThunk


#endif //THUNKS_H_
