/*////////////////////////////////////////////////////////////////////
// timer support classes
// Author : Crusader Mike (mp@appsys.net)
// Description : timer support classes
// Usage :

#include "sf_timer.h"

CTimerHost g_TimerHost;

class CMyClass
{
public:
    CMyClass() : m_Timer(g_TimerHost, this, OnTimer)
    {
        timer.SetTimer(500);
    }

    void Stop()
    {
        timer.KillTimer();
    }

    void OnTimer(DWORD dwTime) {}

protected:
    CTimerThunk<CMyClass>   m_Timer;
};

    History:
    -=-=-=-=
    7/17/2001 -- Initial release
    7/18/2001 -- Some simplifying

////////////////////////////////////////////////////////////////////*/

#ifndef SF_TIMER_H
#define SF_TIMER_H

#define BAD_TIMER_ID    ((UINT)-1)

#pragma warning(disable : 4290)

typedef void (__stdcall *SafeTimerProc)(DWORD dwTime) throw();
class CTimerHost;

#include "Thunks.h"
// include internal timer stuff
#include "sf_help.h"


//////////////////////////////////////////////////////////////////
// CTimerHost -- source of all timers
//               should be used and destroyed only from the thread it was created!!!!

class CTimerHost
{
    CHolder<CReceiver*>     m_vReceiver;
    CHolder<CReceiver*>     m_vRecycle; // receivers, pending to delete
    HWND                    m_hWnd;

public:
    class CError
    {
        DWORD   m_dwError;
    public:
        CError(DWORD dwError = GetLastError()) throw();
        ~CError() throw();
        DWORD Error() const throw();
    };

public:
    CTimerHost() throw(CError);
    ~CTimerHost() throw();

    UINT SetTimer(UINT nElapse, SafeTimerProc pfnCallback) throw();
    void KillTimer(UINT nTimerID) throw();

protected:
    friend class CReceiver;
    void RemoveReceiver(UINT nID) throw();

    //copy protection
    CTimerHost(const CTimerHost&) throw();
    const CTimerHost& operator=(const CTimerHost&) throw();
};


//////////////////////////////////////////////////////////////////
// CTimerThunk<T>

template <class T>
class CTimerThunk : private CAuxThunk<T>
{
	UINT            m_nTimerID;
    CTimerHost&     m_Host;

public:
    typedef void (T::*PFNTimerProc)(DWORD dwTime) throw();

    CTimerThunk(CTimerHost& rTimerHost, T* pThis, PFNTimerProc pfn) throw();
    ~CTimerThunk() throw();

    bool SetTimer(UINT nElapse) throw();
	void KillTimer() throw();
	bool IsRunning() throw();

protected:
    //copy protection
    CTimerThunk(const CTimerThunk&) throw();
    const CTimerThunk& operator=(const CTimerThunk&) throw();
};


/**************************************************************************************
                                 Implementations
**************************************************************************************/


//////////////////////////////////////////////////////////////////
// CTimerHost

inline CTimerHost::CError::CError(DWORD dwError /*= GetLastError()*/) throw()
    :   m_dwError(dwError)
{
}

inline CTimerHost::CError::~CError() throw()
{
}

inline DWORD CTimerHost::CError::Error() const throw()
{
    return m_dwError;
}

inline void CTimerHost::RemoveReceiver(UINT nID) throw()
{
    if (m_vRecycle.size() > nID)
    {
        m_vRecycle.SetAt(nID, NULL);
    }
}


//////////////////////////////////////////////////////////////////
// CTimerThunk

template <class T>
inline CTimerThunk<T>::CTimerThunk(CTimerHost& rTimerHost, T* pThis, PFNTimerProc pfn) throw()
    :   m_nTimerID(BAD_TIMER_ID),
        m_Host(rTimerHost)
{
    InitThunk((TMFP)pfn, pThis);
}

template <class T>
inline CTimerThunk<T>::~CTimerThunk() throw()
{
    if (m_nTimerID != BAD_TIMER_ID)
    {
        m_Host.KillTimer(m_nTimerID);
    }
}

template <class T>
inline bool CTimerThunk<T>::SetTimer(UINT nElapse) throw()
{
    KillTimer();
    m_nTimerID = m_Host.SetTimer(nElapse, (SafeTimerProc)GetThunk());
    return (m_nTimerID != BAD_TIMER_ID);
}

template <class T>
inline void CTimerThunk<T>::KillTimer() throw()
{
    if (m_nTimerID != BAD_TIMER_ID)
    {
        m_Host.KillTimer(m_nTimerID);
        m_nTimerID = BAD_TIMER_ID;
    }
}

template <class T>
inline bool CTimerThunk<T>::IsRunning() throw()
{
    return (m_nTimerID != BAD_TIMER_ID);
}


#endif //SF_TIMER_H
